# Query responses
EXPECTED_GENERATE_TEMPLATE_GQL_RESPONSE_FIELD = "generateCollectorTemplate"
EXPECTED_TTC_RESPONSE_FIELD = "testTelnetConnection"
EXPECTED_TTOC_RESPONSE_FIELD = "testTcpOpenConnection"
EXPECTED_ADD_DC_RESPONSE_FIELD = "createCollectorRecord"

DEFAULT_COLLECTION_REGION = "us-east-1"

# Verbiage
ADD_DC_REGION_PROMPT_VERBIAGE = "Which AWS region would you like to deploy in?"
ADD_DC_PROMPT_VERBIAGE = "Do you want deploy this collector right now? This opens a browser to CF console with a quick create link"



from unittest import TestCase
from unittest.mock import Mock, patch, call

import click
from box import BoxList, Box

from montecarlodata.common.data import MonolithResponse
from montecarlodata.insights.data_insights import InsightsService
from montecarlodata.queries.insights import GET_INSIGHTS, GET_INSIGHT_REPORT
from montecarlodata.utils import GqlWrapper
from tests.test_common_user import _SAMPLE_CONFIG

SAMPLE_RAW = "loki"
SAMPLE_INSIGHT = "key_assets"
SAMPLE_INSIGHT_URL = "https://montecarlodata.com"
SAMPLE_S3_DESTINATION = "s3://bucket/prefix/object"
SAMPLE_FS_DESTINATION = "file://folder1/folder2/object"


class MockRequest:
    raw = SAMPLE_RAW


class InsightServiceTest(TestCase):
    def setUp(self) -> None:
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._service = InsightsService(
            _SAMPLE_CONFIG, request_wrapper=self._request_wrapper_mock
        )

    @patch("montecarlodata.insights.data_insights.click")
    def test_echo_insights(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data=BoxList(
                [
                    {
                        "description": "description",
                        "title": "Key Assets",
                        "available": True,
                    }
                ]
            ),
            errors=None,
        )

        self._service.echo_insights(table_format="plain")
        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=GET_INSIGHTS, operation="getInsights"
        )
        click_mock.echo.assert_called_once_with(
            "Insight     Description    Available\nKey Assets  description.   True"
        )

    def test_get_insight_with_bad_scheme(self):
        with self.assertRaises(click.exceptions.Abort):
            self._service.get_insight(insight=SAMPLE_INSIGHT, destination="foo")

        with self.assertRaises(click.exceptions.Abort):
            self._service.get_insight(insight=SAMPLE_INSIGHT, destination="http://foo")

    @patch.object(InsightsService, "_save_insight_to_disk")
    @patch.object(InsightsService, "_save_insight_to_s3")
    def test_get_insight_routing_s3(self, save_to_s3_mock, save_to_disk_mock):
        self._test_get_insight_routing(destination=SAMPLE_S3_DESTINATION)

        save_to_disk_mock.assert_not_called()
        save_to_s3_mock.assert_called_once_with(
            insight_url=SAMPLE_INSIGHT_URL,
            destination="bucket/prefix/object",
            aws_profile=None,
        )

    @patch.object(InsightsService, "_save_insight_to_disk")
    @patch.object(InsightsService, "_save_insight_to_s3")
    def test_get_insight_routing_fs(self, save_to_s3_mock, save_to_disk_mock):
        self._test_get_insight_routing(destination=SAMPLE_FS_DESTINATION)

        save_to_disk_mock.assert_called_once_with(
            insight_url=SAMPLE_INSIGHT_URL,
            destination="folder1/folder2/object",
            aws_profile=None,
        )
        save_to_s3_mock.assert_not_called()

    @patch("montecarlodata.insights.data_insights.click")
    @patch.object(InsightsService, "_get_insight_url")
    @patch.object(InsightsService, "_save_insight_to_disk")
    @patch.object(InsightsService, "_save_insight_to_s3")
    def test_get_insight_dry(
        self, save_to_s3_mock, save_to_disk_mock, get_mock, mock_click
    ):
        get_mock.return_value = SAMPLE_INSIGHT_URL

        service = InsightsService(
            _SAMPLE_CONFIG, request_wrapper=self._request_wrapper_mock
        )
        service.get_insight(
            insight=SAMPLE_INSIGHT, destination=SAMPLE_FS_DESTINATION, dry=True
        )
        get_mock.assert_called_once_with(insight=SAMPLE_INSIGHT)
        mock_click.echo.assert_called_once_with(SAMPLE_INSIGHT_URL)

        save_to_disk_mock.assert_not_called()
        save_to_s3_mock.assert_not_called()

    @patch("montecarlodata.insights.data_insights.click")
    @patch.object(InsightsService, "_get_insight_url")
    def _test_get_insight_routing(self, get_mock, click_mock, destination: str):
        get_mock.return_value = SAMPLE_INSIGHT_URL

        service = InsightsService(
            _SAMPLE_CONFIG, request_wrapper=self._request_wrapper_mock
        )
        service.get_insight(insight=SAMPLE_INSIGHT, destination=destination)
        get_mock.assert_called_once_with(insight=SAMPLE_INSIGHT)
        click_mock.echo.assert_has_calls(
            [
                call(f"Saving insight to '{destination}'."),
                call("Complete. Have a nice day!"),
            ]
        )
        self.assertEqual(click_mock.echo.call_count, 2)

    def test_get_insights_url(self):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data=Box({"url": SAMPLE_INSIGHT_URL})
        )
        self.assertEqual(
            self._service._get_insight_url(insight=SAMPLE_INSIGHT), SAMPLE_INSIGHT_URL
        )

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=GET_INSIGHT_REPORT,
            operation="getReportUrl",
            variables=dict(
                insight_name=SAMPLE_INSIGHT, report_name=f"{SAMPLE_INSIGHT}.csv"
            ),
        )

    def test_get_insights_url_not_available(self):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data=Box({"url": None})
        )

        with self.assertRaises(click.exceptions.Abort):
            self._service._get_insight_url(insight=SAMPLE_INSIGHT)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=GET_INSIGHT_REPORT,
            operation="getReportUrl",
            variables=dict(
                insight_name=SAMPLE_INSIGHT, report_name=f"{SAMPLE_INSIGHT}.csv"
            ),
        )

    @patch("montecarlodata.insights.data_insights.urllib")
    @patch("montecarlodata.insights.data_insights.mkdirs")
    def test_save_insight_to_disk(self, mkdir_mock, urllib_mock):
        destination = "folder1/folder2/object"
        self._service._save_insight_to_disk(
            insight_url=SAMPLE_INSIGHT_URL, destination=destination
        )
        mkdir_mock.assert_called_once_with("folder1/folder2")
        urllib_mock.request.urlretrieve.assert_called_once_with(
            SAMPLE_INSIGHT_URL, destination
        )

    @patch("montecarlodata.insights.data_insights.urllib")
    @patch("montecarlodata.insights.data_insights.mkdirs")
    def test_save_insight_to_disk_with_no_destination(self, mkdir_mock, urllib_mock):
        with self.assertRaises(click.exceptions.Abort):
            self._service._save_insight_to_disk(
                insight_url=SAMPLE_INSIGHT_URL, destination=None
            )
        mkdir_mock.assert_not_called()
        urllib_mock.request.urlretrieve.assert_not_called()

    @patch("montecarlodata.insights.data_insights.AwsClientWrapper")
    @patch("montecarlodata.insights.data_insights.requests")
    def test_save_insight_to_s3(self, request_mock, utils_mock):
        destination = "bucket/prefix/object"
        request_mock.get.return_value = MockRequest

        self._service._save_insight_to_s3(
            insight_url=SAMPLE_INSIGHT_URL, destination=destination
        )
        request_mock.get.assert_called_once_with(SAMPLE_INSIGHT_URL, stream=True)
        utils_mock().upload_stream_to_s3.assert_called_once_with(
            data=SAMPLE_RAW, bucket="bucket", key="prefix/object"
        )

    @patch("montecarlodata.insights.data_insights.AwsClientWrapper")
    @patch("montecarlodata.insights.data_insights.requests")
    def test_save_insight_to_s3_with_bad_destination(self, request_mock, utils_mock):
        with self.assertRaises(click.exceptions.Abort):
            self._service._save_insight_to_s3(
                insight_url=SAMPLE_INSIGHT_URL, destination="bucket"
            )
        request_mock.assert_not_called()
        utils_mock.assert_not_called()

        request_mock.reset_mock()
        utils_mock.reset_mock()

        with self.assertRaises(click.exceptions.Abort):
            self._service._save_insight_to_s3(
                insight_url=SAMPLE_INSIGHT_URL, destination="bucket/"
            )
        request_mock.assert_not_called()
        utils_mock.assert_not_called()



from montecarlodata.common.common import read_as_base64
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD,
    REDSHIFT_CONNECTION_TYPE,
    REDSHIFT_WAREHOUSE_TYPE,
    SNOWFLAKE_CONNECTION_TYPE,
    SNOWFLAKE_WAREHOUSE_TYPE,
    EXPECTED_SNOWFLAKE_GQL_RESPONSE_FIELD,
    BQ_WAREHOUSE_TYPE,
    EXPECTED_BQ_GQL_RESPONSE_FIELD,
    BQ_CONNECTION_TYPE,
)
from montecarlodata.queries.onboarding import (
    TEST_DATABASE_CRED_MUTATION,
    TEST_SNOWFLAKE_CRED_MUTATION,
    TEST_BQ_CRED_MUTATION,
)


class WarehouseOnboardingService(BaseOnboardingService):
    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_redshift(self, **kwargs) -> None:
        """
        Onboard a redshift connection by validating and adding a connection.
        """
        kwargs["connectionType"] = REDSHIFT_CONNECTION_TYPE
        kwargs["warehouseType"] = REDSHIFT_WAREHOUSE_TYPE
        self.onboard(
            validation_query=TEST_DATABASE_CRED_MUTATION,
            validation_response=EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD,
            connection_type=REDSHIFT_CONNECTION_TYPE,
            **kwargs
        )

    @manage_errors
    def onboard_snowflake(self, **kwargs) -> None:
        """
        Onboard a snowflake connection by validating and adding a connection.
        """
        kwargs["warehouseType"] = SNOWFLAKE_WAREHOUSE_TYPE
        if kwargs.get("private_key"):
            kwargs["private_key"] = read_as_base64(kwargs.pop("private_key")).decode(
                "utf-8"
            )
        self.onboard(
            validation_query=TEST_SNOWFLAKE_CRED_MUTATION,
            validation_response=EXPECTED_SNOWFLAKE_GQL_RESPONSE_FIELD,
            connection_type=SNOWFLAKE_CONNECTION_TYPE,
            **kwargs
        )

    @manage_errors
    def onboard_bq(self, **kwargs) -> None:
        """
        Onboard a BigQuery connection by validating and adding a connection.

        Reads and encodes service file as base64.
        """
        kwargs["warehouseType"] = BQ_WAREHOUSE_TYPE
        kwargs["serviceJson"] = read_as_base64(kwargs.pop("ServiceFile")).decode(
            "utf-8"
        )
        self.onboard(
            validation_query=TEST_BQ_CRED_MUTATION,
            validation_response=EXPECTED_BQ_GQL_RESPONSE_FIELD,
            connection_type=BQ_CONNECTION_TYPE,
            **kwargs
        )



import copy
from unittest import TestCase
from unittest.mock import Mock, patch, call

import click
from box import Box

from montecarlodata.common.data import AwsGlueAthenaResourceProperties
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.discovery.policy_gen import PolicyDiscoveryService
from tests.test_network_discovery import SAMPLE_DC_PROPS

SAMPLE_BUCKET = "foo"
SAMPLE_PATH = "/bar/qux"
SAMPLE_GET_TABLES = {
    "TableList": [
        {"StorageDescriptor": {"Location": f"s3://{SAMPLE_BUCKET}{SAMPLE_PATH}"}}
    ],
    "NextToken": None,
}
SAMPLE_GET_WORKGROUP = {
    "Configuration": {
        "ResultConfiguration": {"OutputLocation": f"s3://{SAMPLE_BUCKET}{SAMPLE_PATH}"}
    }
}


class PolicyDiscoveryTest(TestCase):
    def setUp(self) -> None:
        self.maxDiff = None
        self._cloud_resource_service_mock = Mock(autospec=CloudResourceService)

        self._service = PolicyDiscoveryService(
            cloud_resource_service=self._cloud_resource_service_mock
        )

    @patch("montecarlodata.discovery.policy_gen.click")
    @patch("montecarlodata.discovery.policy_gen.generate_token")
    @patch("montecarlodata.discovery.policy_gen.read_files")
    @patch("montecarlodata.discovery.policy_gen.render_dumped_json")
    def test_generate_cf_role(self, render_mock, read_mock, generate_mock, click_mock):
        account, external_id, file_vals, template_vals = (
            "foo",
            "1234",
            {"foo": "bar"},
            {"template": {"foo": "bar"}},
        )
        files, collector_props = ["a", "b", "c"], Box(
            {"stackArn": f"arn:partition:service:region:{account}:resource"}
        )

        self._cloud_resource_service_mock.get_and_validate_active_collector.return_value = (
            collector_props
        )
        generate_mock.return_value = external_id
        read_mock.return_value = file_vals
        render_mock.return_value = template_vals

        self._service.generate_cf_role(policy_files=files)

        read_mock.assert_called_once_with(files)
        generate_mock.assert_called_once_with()
        render_mock.assert_called_once_with(
            path="../templates/aws_role.json",
            collector_account_id=account,
            collector_external_id=external_id,
            resource_policies=file_vals,
        )
        click_mock.assert_has_calls(
            [
                call.echo(
                    "Generating CloudFormation template for creating a role resource.",
                    err=True,
                ),
                call.echo(template_vals),
            ]
        )

    @patch("montecarlodata.discovery.policy_gen.complain_and_abort")
    def test_generate_cf_role_without_a_policy(self, complain_mock):
        complain_mock.side_effect = click.exceptions.Abort

        with self.assertRaises(click.exceptions.Abort):
            self._service.generate_cf_role(policy_files=[])
        complain_mock.assert_called_once_with("At least one policy is required.")

    @patch.object(PolicyDiscoveryService, "_get_common_glue_athena_props")
    @patch("montecarlodata.discovery.policy_gen.click")
    @patch("montecarlodata.discovery.policy_gen.render_dumped_json")
    def test_generate_glue_policy(self, render_mock, click_mock, common_mock):
        options, database_name, bucket_names, account_id = (
            {"foo": "bar"},
            {"a", "b"},
            {"c", "d"},
            "1234",
        )
        rendering = {"hello": "world"}
        props = copy.deepcopy(SAMPLE_DC_PROPS)

        common_mock.return_value = AwsGlueAthenaResourceProperties(
            database_names=database_name,
            bucket_names=bucket_names,
            dc_resource_props=props,
        )
        props.resources_client.get_caller_identity.return_value = account_id
        render_mock.return_value = rendering

        self._service.generate_glue_policy(**options)

        common_mock.assert_called_once_with(**options)
        render_mock.assert_called_once_with(
            path="../templates/glue_policy.json",
            region="us-east-1",
            account_id=account_id,
            database_names=database_name,
            data_buckets=bucket_names,
        )
        click_mock.assert_has_calls(
            [call.echo("Generating policy for Glue.", err=True), call.echo(rendering)]
        )

    @patch.object(PolicyDiscoveryService, "_get_workgroup_results")
    @patch.object(PolicyDiscoveryService, "_get_common_glue_athena_props")
    @patch("montecarlodata.discovery.policy_gen.click")
    @patch("montecarlodata.discovery.policy_gen.render_dumped_json")
    def test_generate_athena_policy(
        self, render_mock, click_mock, common_mock, workgroup_mock
    ):
        workgroup_name = "workgroup"
        options, bucket_name, path = (
            {"foo": "bar", "workgroup_name": workgroup_name},
            "bucket",
            "prefix/key",
        )
        database_name, bucket_names, account_id, rendering = (
            {"a", "b"},
            {"c", "d"},
            "1234",
            {"hello": "world"},
        )
        props = copy.deepcopy(SAMPLE_DC_PROPS)

        common_mock.return_value = AwsGlueAthenaResourceProperties(
            database_name, bucket_names, props
        )
        workgroup_mock.return_value = (bucket_name, path)
        props.resources_client.get_caller_identity.return_value = account_id
        render_mock.return_value = rendering

        self._service.generate_athena_policy(**options)

        common_mock.assert_called_once_with(foo="bar")
        workgroup_mock.assert_called_once_with(
            workgroup_name=workgroup_name, dc_resource_props=props
        )
        render_mock.assert_called_once_with(
            path="../templates/athena_policy.json",
            region="us-east-1",
            account_id=account_id,
            database_names=database_name,
            data_buckets=bucket_names,
            result_bucket=bucket_name,
            result_path=path,
            workgroup_name=workgroup_name,
        )
        click_mock.assert_has_calls(
            [call.echo("Generating policy for Athena.", err=True), call.echo(rendering)]
        )

    @patch.object(PolicyDiscoveryService, "_get_buckets")
    def test_get_common_glue_athena_props(self, bucket_mock):
        database_names, bucket_names = ["a", "b", "*"], ["c", "d", "d"]
        collector_props, dc_resource_props = Box({"foo": "bar"}), copy.deepcopy(
            SAMPLE_DC_PROPS
        )

        self._cloud_resource_service_mock.get_and_validate_active_collector.return_value = (
            collector_props
        )
        self._cloud_resource_service_mock.get_dc_resource_props.return_value = (
            dc_resource_props
        )

        self.assertEqual(
            self._service._get_common_glue_athena_props(
                database_names=database_names, bucket_names=bucket_names
            ),
            AwsGlueAthenaResourceProperties(
                database_names={"*"},
                bucket_names={"c", "d"},
                dc_resource_props=dc_resource_props,
            ),
        )
        self._cloud_resource_service_mock.get_and_validate_active_collector.assert_called_once_with(
            dc_id=None
        )
        self._cloud_resource_service_mock.get_dc_resource_props.assert_called_once_with(
            collector_props=collector_props,
            get_stack_outputs=False,
            get_stack_params=False,
        )
        bucket_mock.assert_not_called()

    @patch.object(PolicyDiscoveryService, "_get_buckets")
    def test_get_common_glue_athena_props_with_no_buckets(self, bucket_mock):
        database_names, bucket_names, dc_id = ["a", "b", "*"], {"c", "d"}, "1234"
        collector_props, dc_resource_props = Box({"foo": "bar"}), copy.deepcopy(
            SAMPLE_DC_PROPS
        )

        bucket_mock.return_value = bucket_names

        self._cloud_resource_service_mock.get_and_validate_active_collector.return_value = (
            collector_props
        )
        self._cloud_resource_service_mock.get_dc_resource_props.return_value = (
            dc_resource_props
        )

        self.assertEqual(
            self._service._get_common_glue_athena_props(
                database_names=database_names, dc_id=dc_id
            ),
            AwsGlueAthenaResourceProperties(
                database_names={"*"},
                bucket_names=bucket_names,
                dc_resource_props=dc_resource_props,
            ),
        )
        self._cloud_resource_service_mock.get_and_validate_active_collector.assert_called_once_with(
            dc_id=dc_id
        )
        self._cloud_resource_service_mock.get_dc_resource_props.assert_called_once_with(
            collector_props=collector_props,
            get_stack_outputs=False,
            get_stack_params=False,
        )
        bucket_mock.assert_called_once_with(
            database_names={"*"}, dc_resource_props=dc_resource_props
        )

    @patch("montecarlodata.discovery.policy_gen.click.progressbar")
    @patch.object(PolicyDiscoveryService, "_get_table_locations")
    def test_get_buckets(self, table_location_mock, click_mock):
        class MockContext:
            def __enter__(self):
                return self

            def __iter__(self):
                for x in database_names:
                    yield x

            def __exit__(self, *args):
                pass

        tables = Box(SAMPLE_GET_TABLES, camel_killer_box=True)
        database_names, locations = {"foo", "bar"}, {"bucket1", "bucket2"}
        dc_resource_props = copy.deepcopy(SAMPLE_DC_PROPS)

        table_location_mock.return_value = locations
        click_mock.return_value = MockContext()

        dc_resource_props.resources_client.get_glue_tables.return_value = tables

        self.assertEqual(
            self._service._get_buckets(
                database_names=database_names, dc_resource_props=dc_resource_props
            ),
            locations,
        )
        table_location_mock.assert_has_calls(
            [call(tables=tables.table_list), call(tables=tables.table_list)]
        )
        dc_resource_props.resources_client.get_glue_tables.assert_has_calls(
            [
                call(database_name="bar", page_token=None),
                call(database_name="foo", page_token=None),
            ],
            any_order=True,
        )

    def test_handle_wildcards(self):
        self.assertEqual(self._service._handle_wildcards(None), set())
        self.assertEqual(self._service._handle_wildcards([]), set())
        self.assertEqual(
            self._service._handle_wildcards(["foo", "foo", "foo"]), {"foo"}
        )
        self.assertEqual(self._service._handle_wildcards(["foo", "foo", "*"]), {"*"})

    @patch("montecarlodata.discovery.policy_gen.click.progressbar")
    def test_get_workgroup_results(self, click_mock):
        workgroup_name, dc_resource_props = "foo", copy.deepcopy(SAMPLE_DC_PROPS)
        workgroup_details = Box(SAMPLE_GET_WORKGROUP, camel_killer_box=True)

        dc_resource_props.resources_client.get_athena_workgroup.return_value = (
            workgroup_details
        )
        self.assertEqual(
            self._service._get_workgroup_results(
                workgroup_name, dc_resource_props=dc_resource_props
            ),
            (SAMPLE_BUCKET, SAMPLE_PATH),
        )
        dc_resource_props.resources_client.get_athena_workgroup.assert_called_once_with(
            workgroup=workgroup_name
        )

    def test_get_table_locations(self):
        tables = Box(SAMPLE_GET_TABLES, camel_killer_box=True)
        self.assertEqual(
            list(self._service._get_table_locations(tables=tables.table_list)),
            [SAMPLE_BUCKET],
        )

    def test_get_table_locations_with_invalid_schemes(self):
        tables = Box(SAMPLE_GET_TABLES, camel_killer_box=True)
        tables.table_list[0].storage_descriptor.location = "invalid"
        self.assertEqual(
            list(self._service._get_table_locations(tables=tables.table_list)), []
        )



from typing import Any, Dict, Optional
from unittest import TestCase
from unittest.mock import Mock, patch

from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.data_lake.hive import HiveOnboardingService
from montecarlodata.integrations.onboarding.fields import AWS_RDS_CA_CERT
from montecarlodata.queries.onboarding import (
    TEST_DATABASE_CRED_MUTATION,
    TEST_S3_CRED_MUTATION,
    TEST_HIVE_SQL_CRED_MUTATION,
)
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.test_base_onboarding import _SAMPLE_BASE_OPTIONS
from tests.test_common_user import _SAMPLE_CONFIG


class HiveOnboardingTest(TestCase):
    _DEFAULT_JOB_LIMITS = {
        "get_partition_locations": True,
        "max_partition_locations": 50,
    }

    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = HiveOnboardingService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock,
        )

    def test_hive_mysql_flow_without_a_catalog(self):
        sample_input = {"use_ssl": True, "catalog": None, **_SAMPLE_BASE_OPTIONS}
        self._test_hive_mysql_flow(
            sample_input=sample_input, job_limits=self._DEFAULT_JOB_LIMITS
        )

    def test_hive_mysql_flow_with_a_catalog(self):
        sample_input = {"use_ssl": True, "catalog": "foo", **_SAMPLE_BASE_OPTIONS}
        self._test_hive_mysql_flow(
            sample_input=sample_input,
            job_limits={"catalog_name": "foo", **self._DEFAULT_JOB_LIMITS},
        )

    @patch.object(HiveOnboardingService, "onboard")
    def test_hive_s3_flow(self, onboard_mock):
        expected_options = {**_SAMPLE_BASE_OPTIONS, **{"connectionType": "hive-s3"}}

        self._service.onboard_hive_s3(**_SAMPLE_BASE_OPTIONS)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_S3_CRED_MUTATION,
            validation_response="testS3Credentials",
            connection_type="hive-s3",
            job_types=["query_logs"],
            **expected_options
        )

    @patch.object(HiveOnboardingService, "onboard")
    def test_hive_sql_flow(self, onboard_mock):
        self._service.onboard_hive_sql(**_SAMPLE_BASE_OPTIONS)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_HIVE_SQL_CRED_MUTATION,
            validation_response="testHiveCredentials",
            connection_type="hive",
            **_SAMPLE_BASE_OPTIONS
        )

    @patch.object(HiveOnboardingService, "onboard")
    def _test_hive_mysql_flow(
        self, onboard_mock: Any, sample_input: Dict, job_limits: Optional[Dict] = None
    ):
        expected_options = {
            "ssl_options": {"ca": AWS_RDS_CA_CERT},
            "connectionType": "hive-mysql",
            **_SAMPLE_BASE_OPTIONS,
        }

        self._service.onboard_hive_mysql(**sample_input)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_DATABASE_CRED_MUTATION,
            validation_response="testDatabaseCredentials",
            connection_type="hive-mysql",
            job_limits=job_limits,
            **expected_options
        )



from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD,
    EXPECTED_HIVE_S3_GQL_RESPONSE_FIELD,
    EXPECTED_HIVE_SQL_GQL_RESPONSE_FIELD,
    HIVE_MYSQL_CONNECTION_TYPE,
    HIVE_S3_CONNECTION_TYPE,
    HIVE_SQL_CONNECTION_TYPE,
    QL_JOB_TYPE,
    AWS_RDS_CA_CERT,
    PRESTO_CATALOG_KEY,
    HIVE_GET_PARTS_KEY,
    HIVE_MAX_PARTS_KEY,
    HIVE_MAX_PARTS_DEFAULT_VALUE,
)
from montecarlodata.queries.onboarding import (
    TEST_DATABASE_CRED_MUTATION,
    TEST_S3_CRED_MUTATION,
    TEST_HIVE_SQL_CRED_MUTATION,
)


class HiveOnboardingService(BaseOnboardingService):
    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_hive_mysql(self, **kwargs) -> None:
        """
        Onboard a hive-mysql connection by validating and adding a connection
        """
        if kwargs["use_ssl"]:
            kwargs["ssl_options"] = {"ca": AWS_RDS_CA_CERT}
            kwargs.pop("use_ssl")

        catalog = kwargs.pop("catalog", None)
        job_limits = {
            HIVE_GET_PARTS_KEY: True,
            HIVE_MAX_PARTS_KEY: HIVE_MAX_PARTS_DEFAULT_VALUE,
        }

        if catalog:
            job_limits[PRESTO_CATALOG_KEY] = catalog

        kwargs["connectionType"] = HIVE_MYSQL_CONNECTION_TYPE
        self.onboard(
            validation_query=TEST_DATABASE_CRED_MUTATION,
            validation_response=EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD,
            connection_type=HIVE_MYSQL_CONNECTION_TYPE,
            job_limits=job_limits,
            **kwargs
        )

    @manage_errors
    def onboard_hive_s3(self, **kwargs) -> None:
        """
        Onboard a presto-s3 connection by validating and adding a connection
        """
        kwargs["connectionType"] = HIVE_S3_CONNECTION_TYPE
        self.onboard(
            validation_query=TEST_S3_CRED_MUTATION,
            validation_response=EXPECTED_HIVE_S3_GQL_RESPONSE_FIELD,
            connection_type=HIVE_S3_CONNECTION_TYPE,
            job_types=QL_JOB_TYPE,
            **kwargs
        )

    @manage_errors
    def onboard_hive_sql(self, **kwargs) -> None:
        """
        Onboard a hive-sql connection by validating and adding a connection
        """
        self.onboard(
            validation_query=TEST_HIVE_SQL_CRED_MUTATION,
            validation_response=EXPECTED_HIVE_SQL_GQL_RESPONSE_FIELD,
            connection_type=HIVE_SQL_CONNECTION_TYPE,
            **kwargs
        )



from unittest import TestCase
from unittest.mock import patch, call

import click

from montecarlodata.common.data import MonolithResponse
from montecarlodata.errors import (
    echo_error,
    complain_and_abort,
    manage_errors,
    abort_on_gql_errors,
    prompt_connection,
)
from tests.helpers import capture_function


class HasErrors:
    def __init__(self):
        self._abort_on_error = True

    @manage_errors
    def function_with_errors(self):
        raise ZeroDivisionError


class ErrorTest(TestCase):
    @patch("montecarlodata.errors.click")
    def test_echo_error(self, click_mock):
        message = "foo"
        echo_error(message=message)
        click_mock.assert_has_calls([call.echo(f"Error - {message}", err=True)])

    def test_complain_and_abort(self):
        with self.assertRaises(click.exceptions.Abort):
            complain_and_abort("test")

    def test_manage_errors(self):
        with self.assertRaises(click.exceptions.Abort):
            HasErrors().function_with_errors()

    @patch("montecarlodata.errors.click")
    def test_abort_on_gql_errors(self, click_mock):
        message = "foo"
        click_mock.echo.side_effect = click.Abort()

        with self.assertRaises(click.exceptions.Abort):
            abort_on_gql_errors(
                response=MonolithResponse(errors=[{"message": message}])
            )
        click_mock.assert_has_calls([call.echo("Error - foo", err=True)])

    @patch("montecarlodata.errors.click")
    def test_prompt_connection(self, click_mock):
        message = "Hello, world!"
        prompt_connection(message=message, skip_prompt=False)
        click_mock.confirm.assert_called_once_with(message, abort=True)

    @patch("montecarlodata.errors.click")
    def test_prompt_connection_with_decline(self, click_mock):
        click_mock.confirm.side_effect = click.Abort

        with self.assertRaises(click.exceptions.Abort):
            prompt_connection(message="Hello, world!", skip_prompt=False)



from typing import Optional, List, Dict

import click
from tabulate import tabulate

from montecarlodata.common.common import normalize_gql
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.fields import GQL_TO_FRIENDLY_CONNECTION_MAP


class OnboardingStatusService:
    _INTEGRATION_FRIENDLY_HEADERS = [
        "Integration",
        "Name",
        "ID",
        "Connection",
        "Created on (UTC)",
    ]

    def __init__(self, config: Config, user_service: Optional[UserService] = None):
        self._abort_on_error = True
        self._user_service = user_service or UserService(config=config)

    @manage_errors
    def display_integrations(
        self,
        headers: Optional[str] = "firstrow",
        table_format: Optional[str] = "fancy_grid",
    ):
        """
        Display active integrations in an easy to read table. E.g.
        ╒═══════════════════╤═══════════╤══════════════════════════════════════╤═════════════════════════════╤══════════════════════════════════╕
        │ Integration       │ Name      │ ID                                   │ Connection                  │ Created on (UTC)                 │
        ╞═══════════════════╪═══════════╪══════════════════════════════════════╪═════════════════════════════╪══════════════════════════════════╡
        │ Snowflake         │ rick_test │ 9403e385-1a56-4c3b-b185-4f2bb8e324d3 │ account: hda34492.us-east-1 │ 2022-03-25T19:55:11.858022+00:00 │
        ╘═══════════════════╧═══════════╧══════════════════════════════════════╧═════════════════════════════╧══════════════════════════════════╛
        """
        table = [self._INTEGRATION_FRIENDLY_HEADERS]
        for integration in (self._user_service.warehouses or [{}]) + (
            self._user_service.bi_containers or [{}]
        ):
            table += self._build_table_record(
                name=integration.get("name"), connections=integration.get("connections")
            )
        click.echo(tabulate(table, headers=headers, tablefmt=table_format))

    def _build_table_record(
        self, name: str, connections: List[Dict]
    ) -> List[List[str]]:
        table = []
        for connection in connections or []:
            table.append(
                [
                    GQL_TO_FRIENDLY_CONNECTION_MAP.get(
                        normalize_gql(connection["type"]), connection["type"]
                    ),
                    name,
                    connection["uuid"],
                    self._build_connection_info(connection),
                    connection["createdOn"],
                ]
            )  # order by the friendly headers, defaulting to gql response if not found
        return table

    @staticmethod
    def _build_connection_info(connection: Dict) -> Optional[str]:
        identifiers = connection.get("connectionIdentifiers")
        if not identifiers:
            return None
        connection_info = []
        for identifier in identifiers:
            key = identifier.get("key")
            value = identifier.get("value")
            connection_info.append(f"{key}: {value}")
        return "\n".join(connection_info)



import json
import re
from typing import Dict, Optional, List, Iterable, Any

import boto3
import requests
from box import Box, BoxList
from retry import retry

import montecarlodata.settings as settings
from montecarlodata.config import Config
from montecarlodata.common.common import boxify
from montecarlodata.common.data import MonolithResponse
from montecarlodata.errors import manage_errors, complain_and_abort, abort_on_gql_errors


class Wrapper:
    def __init__(
        self,
        abort_on_error: Optional[bool] = True,
        disable_handle_errors: Optional[bool] = False,
    ):
        self._abort_on_error = abort_on_error
        self._disable_handle_errors = disable_handle_errors


class GqlWrapper(Wrapper):
    def __init__(self, config: Config, **kwargs):
        self._mcd_id = config.mcd_id
        self._mcd_token = config.mcd_token
        self._endpoint = config.mcd_api_endpoint

        super().__init__(**kwargs)

    def make_request(
        self, query: str, variables: Optional[Dict] = None
    ) -> Optional[Dict]:
        """
        Make a GraphQl request to the MCD API. Aborts on error if set.
        """
        response = self._make_request(query=query, variables=variables)
        if response:
            if self._abort_on_error and response.get("errors"):
                complain_and_abort(response["errors"])
            return response["data"]

    def make_request_v2(
        self, query: str, operation: str, variables: Optional[Dict] = None
    ) -> MonolithResponse:
        """
        Make a GraphQl request to the MCD API and parse out the results into data and errors. Aborts on error if set.
        """
        response = self._make_request(
            query=query,
            variables=self.convert_snakes_to_camels(variables)
            if variables
            else variables,
        )
        data = response.get("data", {}).get(operation)
        if data:
            try:
                data = Box(data)
            except ValueError:
                data = BoxList(data)
        parsed_response = MonolithResponse(data=data, errors=response.get("errors"))

        if self._abort_on_error:
            abort_on_gql_errors(parsed_response)
        return parsed_response

    def convert_snakes_to_camels(self, dict_: Dict) -> Dict:
        """
        Converts dictionary keys from snake_case to camelCase as Gql is very opinionated
        """
        converted = {}
        for k, v in dict_.items():
            k = re.sub(r"_([a-z])", lambda x: x.group(1).upper(), k)
            if isinstance(v, Dict):
                v = self.convert_snakes_to_camels(v)
            converted[k] = v
        return converted

    @manage_errors
    def _make_request(
        self, query: str, variables: Optional[Dict] = None
    ) -> Optional[Dict]:
        headers = {
            "x-mcd-id": self._mcd_id,
            "x-mcd-token": self._mcd_token,
            "Content-Type": "application/json",
        }
        payload = {"query": query, "variables": variables or {}}

        if settings.MCD_USER_ID_HEADER:
            headers = {
                "user-id": settings.MCD_USER_ID_HEADER
            }  # override for local development

        response = self._post(headers=headers, payload=payload)
        response.raise_for_status()

        return json.loads(response.text)

    @retry(tries=3, delay=0.2, backoff=2, max_delay=1)
    def _post(self, headers: Dict, payload: Dict) -> requests.request:
        return requests.post(self._endpoint, json=payload, headers=headers)


class AwsClientWrapper(Wrapper):
    _CF_CREATE_WAITER = "stack_create_complete"
    _CF_UPDATE_WAITER = "stack_update_complete"
    _DEFAULT_WAITER_CONFIG = {"Delay": 5, "MaxAttempts": 720}
    _DEFAULT_CAPABILITIES = "CAPABILITY_IAM"
    _STACK_CREATE_STATE = "CREATE_COMPLETE"
    _STACK_UPDATE_STATE = "UPDATE_COMPLETE"
    # Map from the shorthand state form available in the AWS CLI to the actual states.
    _EMR_STATES = {
        "active": ["STARTING", "BOOTSTRAPPING", "RUNNING", "WAITING", "TERMINATING"],
        "terminated": ["TERMINATED"],
        "failed": ["TERMINATED_WITH_ERRORS"],
    }

    def __init__(
        self,
        profile_name: Optional[str] = None,
        region_name: Optional[str] = None,
        session: Optional[boto3.session.Session] = None,
    ):
        super().__init__()

        self._session = session or self._get_session(
            profile_name=profile_name, region_name=region_name
        )
        self._test_if_region_is_set()

    @property
    def region(self):
        return self._session.region_name

    @manage_errors
    def get_stack_details(self, stack_id: str) -> Dict:
        """
        Retrieve stack details (description) from DC
        """
        return self._session.client("cloudformation").describe_stacks(
            StackName=stack_id
        )

    @manage_errors
    @boxify()
    def get_stack_outputs(self, stack_id: str) -> Box:
        """
        Convenience utility to retrieve stack outputs from DC
        """
        return self.get_stack_details(stack_id)["Stacks"][0]["Outputs"]

    @manage_errors
    def get_stack_parameters(self, stack_id: str) -> List[Dict]:
        """
        Convenience utility to retrieve stack parameters from the DC
        """
        return self.get_stack_details(stack_id)["Stacks"][0].get("Parameters")

    @manage_errors
    def create_stack(
        self,
        stack_name: str,
        template_link: str,
        termination_protection: bool,
        parameters: Optional[List[Dict]] = None,
    ) -> bool:
        """
        Attempts to deploy the DC stack - returning True on success
        """
        deployment_options = dict(
            StackName=stack_name,
            TemplateURL=template_link,
            Capabilities=[self._DEFAULT_CAPABILITIES],
            EnableTerminationProtection=termination_protection,
        )
        if parameters:
            deployment_options["Parameters"] = parameters

        response = self._session.client("cloudformation").create_stack(
            **deployment_options
        )
        stack_id = response["StackId"]
        return self._wait_on_stacks(
            stack_id=stack_id,
            waiter=self._CF_CREATE_WAITER,
            success_state=self._STACK_CREATE_STATE,
        )

    @manage_errors
    def upgrade_stack(
        self, stack_id: str, template_link: str, parameters: List[Dict]
    ) -> bool:
        """
        Attempts to upgrade the DC stack - returning True on success
        """
        self._session.client("cloudformation").update_stack(
            StackName=stack_id,
            TemplateURL=template_link,
            Capabilities=[self._DEFAULT_CAPABILITIES],
            Parameters=parameters,
        )
        return self._wait_on_stacks(
            stack_id=stack_id,
            waiter=self._CF_UPDATE_WAITER,
            success_state=self._STACK_UPDATE_STATE,
        )

    @manage_errors
    def deploy_gateway(self, gateway_id: str, stage: Optional[str] = "prod") -> None:
        """
        Attempts to deploy the DC gateway
        """
        self._session.client("apigateway").create_deployment(
            restApiId=gateway_id, stageName=stage
        )

    @manage_errors
    def upload_file(self, bucket_name: str, object_name: str, file_path: str) -> None:
        """
        Upload a file to s3://bucket/object from file path
        """
        self._session.client("s3").upload_file(file_path, bucket_name, object_name)

    @manage_errors
    def create_role(
        self, role_name: str, trust_policy: str, tags: Optional[List] = None
    ) -> str:
        """
        Creates an IAM role and returns the ARN on success
        """
        return self._session.client("iam").create_role(
            RoleName=role_name, AssumeRolePolicyDocument=trust_policy, Tags=tags
        )["Role"]["Arn"]

    @manage_errors
    def attach_inline_policy(
        self, role_name: str, policy_name: str, policy_doc: str
    ) -> str:
        """
        Attaches an inline policy to an existing role
        """
        return self._session.client("iam").put_role_policy(
            RoleName=role_name, PolicyName=policy_name, PolicyDocument=policy_doc
        )

    @manage_errors
    def get_emr_cluster_details(
        self, created_after: Optional[str] = None, states: Optional[List] = None
    ) -> Iterable[Dict]:
        """
        Retrieves details of all EMR clusters
        """
        client = self._session.client("emr")
        params = {}
        if created_after:
            params["CreatedAfter"] = created_after
        if states:
            params["ClusterStates"] = [
                mapped for state in states for mapped in self._EMR_STATES[state]
            ]
        pages = client.get_paginator("list_clusters").paginate(**params)
        for page in pages:
            for cluster in page["Clusters"]:
                yield client.describe_cluster(ClusterId=cluster["Id"]).get("Cluster")

    def get_bucket_location(self, bucket_name: str) -> str:
        """
        Get the location of a bucket
        """
        region = self._session.client("s3").get_bucket_location(Bucket=bucket_name)[
            "LocationConstraint"
        ]
        if region is None:
            return (
                settings.DEFAULT_AWS_REGION
            )  # Buckets in us-east-1 have a LocationConstraint of null
        return region

    def get_caller_identity(self) -> str:
        """
        Get the Account ID of the caller
        """
        return self._session.client("sts").get_caller_identity()["Account"]

    def create_sns_topic(self, name: str) -> str:
        """
        Create an SNS topic in the account
        """
        return self._session.client("sns").create_topic(Name=name)[
            "TopicArn"
        ]  # This action is idempotent

    @boxify()
    def get_topic_attributes(self, arn: str) -> Box:
        """
        Get attributes for a topic
        """
        return self._session.client("sns").get_topic_attributes(TopicArn=arn)

    def set_topic_attributes(self, arn: str, name: str, value: str) -> None:
        """
        Set attributes for a topic
        """
        return self._session.client("sns").set_topic_attributes(
            TopicArn=arn, AttributeName=name, AttributeValue=value
        )

    @boxify()
    def get_q_attributes(self, name: str, attributes: List[str]) -> Box:
        """
        Get attributes for a sqs
        """
        return self._session.client("sqs").get_queue_attributes(
            QueueUrl=self.get_queue_url(name), AttributeNames=attributes
        )

    def set_q_attributes(self, name: str, attributes: List[str]) -> Box:
        """
        Set attributes for a sqs
        """
        return self._session.client("sqs").set_queue_attributes(
            QueueUrl=self.get_queue_url(name), Attributes=attributes
        )

    def get_queue_url(self, name: str) -> str:
        """
        Get queue url from name
        """
        return self._session.client("sqs").get_queue_url(QueueName=name)["QueueUrl"]

    @boxify()
    def get_bucket_event_config(
        self,
        name: str,
    ) -> Box:
        """
        Get existing bucket config
        """
        return self._session.client("s3").get_bucket_notification_configuration(
            Bucket=name
        )

    def set_bucket_event_config(self, name: str, notification_config: Dict) -> None:
        """
        Update bucket config
        """
        return self._session.client("s3").put_bucket_notification_configuration(
            Bucket=name, NotificationConfiguration=notification_config
        )

    def list_topic_subscriptions(self, arn: str) -> List:
        """
        List subscriptions for a topic
        """
        for page in (
            self._session.client("sns")
            .get_paginator("list_subscriptions_by_topic")
            .paginate(TopicArn=arn)
        ):
            if page and page.get("Subscriptions"):
                yield page["Subscriptions"]

    def subscribe_to_topic(
        self,
        arn: str,
        endpoint: str,
        protocol: Optional[str] = "sqs",
        attributes: Optional[Dict] = None,
    ) -> None:
        """
        Subscribe to a topic
        """
        self._session.client("sns").subscribe(
            TopicArn=arn, Endpoint=endpoint, Protocol=protocol, Attributes=attributes
        )

    def upload_stream_to_s3(self, bucket: str, key: str, data: Any):
        """
        Uploads a file-like object to S3
        """
        self._session.client("s3").upload_fileobj(data, bucket, key)

    @manage_errors
    def _test_if_region_is_set(self) -> None:
        """
        Test if region is set by creating a client. Abort on failure
        """
        self._session.client("cloudformation")  # throws NoRegion if not set

    @manage_errors
    def _get_session(
        self, profile_name: str, region_name: str
    ) -> Optional[boto3.session.Session]:
        """
        Create a session using named profile/region if set. Uses AWS defaults if not set.
        """
        session_args = (("profile_name", profile_name), ("region_name", region_name))
        return boto3.session.Session(
            **{k: v for k, v in session_args if v}
        )  # throws ProfileNotFound if invalid

    def _wait_on_stacks(self, stack_id: str, waiter: str, success_state: str) -> bool:
        """
        Wait for stack to finish operation
        """
        self._session.client("cloudformation").get_waiter(waiter).wait(
            StackName=stack_id, WaiterConfig=self._DEFAULT_WAITER_CONFIG
        )
        return (
            self.get_stack_details(stack_id)["Stacks"][0]["StackStatus"]
            == success_state
        )

    @boxify(use_snakes=True, default_box_attr=None, default_box=True)
    def describe_redshift_cluster(self, cluster_id: str) -> Box:
        return self._session.client("redshift").describe_clusters(
            ClusterIdentifier=cluster_id
        )["Clusters"][0]

    @boxify(use_snakes=True)
    def describe_cluster_subnet_groups(self, subnet_group: str) -> Box:
        return self._session.client("redshift").describe_cluster_subnet_groups(
            ClusterSubnetGroupName=subnet_group
        )["ClusterSubnetGroups"][0]

    @boxify(use_snakes=True)
    def describe_vpc(self, vpc_id: str) -> Box:
        return self._session.client("ec2").describe_vpcs(VpcIds=[vpc_id])["Vpcs"][0]

    @boxify(use_snakes=True)
    def describe_routes(
        self,
        filter_vals: List[str],
        filter_key: Optional[str] = "association.subnet-id",
    ) -> Box:
        return self._session.client("ec2").describe_route_tables(
            Filters=[{"Name": filter_key, "Values": filter_vals}]
        )

    @boxify(use_snakes=True, default_box_attr=None, default_box=True)
    def get_glue_tables(self, database_name: str, page_token: Optional[str] = None):
        return self._session.client("glue").get_tables(
            DatabaseName=database_name,
            **(dict(NextToken=page_token) if page_token else {})
        )

    @boxify(use_snakes=True, default_box_attr=None, default_box=True)
    def get_athena_workgroup(self, workgroup: str):
        return self._session.client("athena").get_work_group(WorkGroup=workgroup)[
            "WorkGroup"
        ]



import sys
import click

import montecarlodata.settings as settings
from montecarlodata.management.commands import management
from montecarlodata.collector.commands import collectors
from montecarlodata.common.user import UserService
from montecarlodata.config import ConfigManager
from montecarlodata.dataimport.commands import import_subcommand
from montecarlodata.discovery.commands import discovery
from montecarlodata.iac.commands import monitors
from montecarlodata.insights.commands import insights
from montecarlodata.integrations.commands import integrations
from montecarlodata.tools import dump_help


@click.group(help="Monte Carlo's CLI.")
@click.option(
    "--profile",
    default=settings.DEFAULT_PROFILE_NAME,
    help="Specify an MCD profile name. Uses default otherwise.",
)
@click.option(
    "--config-path",
    default=settings.DEFAULT_CONFIG_PATH,
    type=click.Path(dir_okay=True),
    help=f"Specify path where to look for config file. Uses {settings.DEFAULT_CONFIG_PATH} otherwise.",
)
@click.version_option()
@click.pass_context
def entry_point(ctx, profile, config_path):
    """
    Entry point for all subcommands and options. Reads configuration and sets as context, except when configuring
    or getting help.
    """
    if (
        ctx.invoked_subcommand != settings.CONFIG_SUB_COMMAND
        and settings.HELP_FLAG not in sys.argv[1:]
    ):
        config = ConfigManager(profile_name=profile, base_path=config_path).read()
        if not config:
            ctx.abort()
        ctx.obj = {"config": config}


@click.command(help="Configure the CLI.")
@click.option(
    "--profile-name",
    required=False,
    help="Specify a profile name for configuration.",
    default=settings.DEFAULT_PROFILE_NAME,
)
@click.option(
    "--config-path",
    required=False,
    help="Specify path where to look for config file.",
    default=settings.DEFAULT_CONFIG_PATH,
    type=click.Path(dir_okay=True),
)
@click.option("--mcd-id", prompt="Key ID", help="Monte Carlo token user ID.")
@click.option(
    "--mcd-token", prompt="Secret", help="Monte Carlo token value.", hide_input=True
)
@click.option(
    "--aws-profile",
    prompt="AWS profile name",
    help="AWS named profile with access to the DC stack and resources (e.g. CloudFormation and S3).",
    default="",
)
@click.option(
    "--aws-region",
    prompt="AWS region",
    help="AWS region where DC is deployed.",
    default=settings.DEFAULT_AWS_REGION,
)
def configure(profile_name, config_path, mcd_id, mcd_token, aws_profile, aws_region):
    """
    Special subcommand for configuring the CLI
    """
    ConfigManager(profile_name=profile_name, base_path=config_path).write(
        mcd_id=mcd_id,
        mcd_token=mcd_token,
        aws_profile=aws_profile,
        aws_region=aws_region,
    )


@click.command(help="Validate that the CLI can Connect to Monte Carlo.")
@click.pass_obj
def validate(ctx):
    """
    Special subcommand for validating the CLI was correctly configured
    """
    click.echo(f"Hi, {UserService(config=ctx['config']).user.first_name}! All is well.")


@entry_point.command(help="Echo all help text.", name="help")
def echo_help():
    """
    Special subcommand to echo all help text.
    """
    dump_help(entry_point)


entry_point.add_command(integrations)
entry_point.add_command(configure)
entry_point.add_command(validate)
entry_point.add_command(collectors)
entry_point.add_command(discovery)
entry_point.add_command(monitors)
entry_point.add_command(import_subcommand)
entry_point.add_command(insights)
entry_point.add_command(management)

# to allow this to be run as a script within an IDE (for debugging)
if __name__ == "__main__":
    entry_point()



import click
from enum import Enum
from montecarlodata.common.data import ConnectionType
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.queries.keys import Queries
from montecarlodata.utils import GqlWrapper
from tabulate import tabulate
from typing import Dict, List, Optional


class IntegrationKeyScope(Enum):
    Spark = "spark"
    DatabricksMetadata = "databricksmetadata"

    @classmethod
    def values(cls) -> List[str]:
        return list(map(lambda s: s.name, cls))


class IntegrationKeyService:

    _table_headers = ["Id", "Description", "Scope", "Created", "Created By"]

    def __init__(
        self,
        config: Config,
        gql: Optional[GqlWrapper] = None,
        user_service: Optional[UserService] = None,
    ):

        self._gql = gql or GqlWrapper(config)
        self._user_service = user_service or UserService(config=config)

        # used by @manage_errors decorator
        self._abort_on_error = True

    @manage_errors
    def create(self, description: str, scope: str):
        response = self._gql.make_request_v2(
            query=Queries.create.query,
            variables=self._resolve_variables(description, scope),
            operation=Queries.create.operation,
        )

        click.echo(f"Key id: {response.data.key.id}")
        click.echo(f"Key secret: {response.data.key.secret}")

    def _resolve_variables(self, description: str, scope: str) -> Dict:
        variables = {"description": description, "scope": scope}

        if scope.lower() == IntegrationKeyScope.Spark.value:
            variables["warehouseIds"] = [self._resolve_lake_warehouse_id()]

        return variables

    def _resolve_lake_warehouse_id(self) -> str:
        lake_warehouse_ids = [
            w["uuid"]
            for w in self._user_service.warehouses
            if w["connectionType"] == ConnectionType.DataLake.value
        ]

        num_lakes = len(lake_warehouse_ids)
        if num_lakes == 0:
            complain_and_abort(
                "Unable to resolve data lake connection: no lake connection found."
            )
        elif num_lakes > 1:
            complain_and_abort(
                "Unable to resolve data lake connection: multiple lake connections found."
            )
        else:
            return lake_warehouse_ids[0]

    @manage_errors
    def delete(self, key_id: str):
        response = self._gql.make_request_v2(
            query=Queries.delete.query,
            variables={"keyId": key_id},
            operation=Queries.delete.operation,
        )

        if response.data.deleted:
            click.echo("Key has been deleted.")
        else:
            click.echo("Key was not deleted.")

    @manage_errors
    def get_all(self, table_format: Optional[str] = "fancy_grid"):
        response = self._gql.make_request_v2(
            query=Queries.get_all.query, operation=Queries.get_all.operation
        )

        data = [
            [
                key.id,
                key.description,
                key.scope,
                key.createdTime,
                f"{key.createdBy.firstName} {key.createdBy.lastName}",
            ]
            for key in response.data
        ]

        table = tabulate(data, headers=self._table_headers, tablefmt=table_format)
        click.echo(table)



import click
from box import BoxList, Box
from montecarlodata.common.data import MonolithResponse
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.integrations.keys import (
    IntegrationKeyScope,
    IntegrationKeyService,
    Queries,
)
from montecarlodata.utils import GqlWrapper
from unittest import TestCase
from unittest.mock import Mock, patch


class IntegrationKeyServiceTests(TestCase):

    config = Config(
        mcd_id="12345",
        mcd_token="67890",
        mcd_api_endpoint="https://api.getmontecarlo.com/graphql",
    )

    def setUp(self) -> None:
        self._request_wrapper = Mock(autospec=GqlWrapper)
        self._user_service = Mock(autospec=UserService)
        self._service = IntegrationKeyService(
            config=self.config,
            gql=self._request_wrapper,
            user_service=self._user_service,
        )

    @patch.object(click, "echo")
    def test_create_spark_key(self, echo: Mock):
        # given
        description = "test key"
        scope = "spark"
        warehouses = [
            {"uuid": "lake", "connectionType": "DATA_LAKE"},
            {"uuid": "snowflake", "connectionType": "SNOWFLAKE"},
        ]
        key = {"id": "key-id", "secret": "key-secret"}

        self._user_service.warehouses = warehouses
        self._request_wrapper.make_request_v2.return_value = MonolithResponse(
            data=Box({"key": key}), errors=None
        )

        # when
        self._service.create(description=description, scope=scope)

        # then
        self._request_wrapper.make_request_v2.assert_called_once_with(
            query=Queries.create.query,
            variables={
                "description": description,
                "scope": scope,
                "warehouseIds": [warehouses[0]["uuid"]],
            },
            operation=Queries.create.operation,
        )

        echo_calls = echo.call_args_list
        self.assertEqual(2, len(echo_calls))
        self.assertEqual(f"Key id: {key['id']}", echo_calls[0][0][0])
        self.assertEqual(f"Key secret: {key['secret']}", echo_calls[1][0][0])

    @patch.object(click, "echo")
    def test_create_spark_key_with_multiple_lakes(self, echo: Mock):
        # given
        description = "test key"
        scope = "spark"
        warehouses = [
            {"uuid": "lake-1", "connectionType": "DATA_LAKE"},
            {"uuid": "lake-2", "connectionType": "DATA_LAKE"},
        ]
        key = {"id": "key-id", "secret": "key-secret"}

        self._user_service.warehouses = warehouses
        self._request_wrapper.make_request_v2.return_value = MonolithResponse(
            data=Box({"key": key}), errors=None
        )

        # when
        with self.assertRaises(click.Abort):
            self._service.create(description=description, scope=scope)

        # then
        self._request_wrapper.make_request_v2.assert_not_called()
        echo.assert_called_once_with(
            "Error - Unable to resolve data lake connection: multiple lake connections found.",
            err=True,
        )

    @patch.object(click, "echo")
    def test_create_spark_key_with_no_lake(self, echo: Mock):
        # given
        description = "test key"
        scope = "spark"
        warehouses = [{"uuid": "snowflake", "connectionType": "SNOWFLAKE"}]
        key = {"id": "key-id", "secret": "key-secret"}

        self._user_service.warehouses = warehouses
        self._request_wrapper.make_request_v2.return_value = MonolithResponse(
            data=Box({"key": key}), errors=None
        )

        # when
        with self.assertRaises(click.Abort):
            self._service.create(description=description, scope=scope)

        # then
        self._request_wrapper.make_request_v2.assert_not_called()
        echo.assert_called_once_with(
            "Error - Unable to resolve data lake connection: no lake connection found.",
            err=True,
        )

    @patch.object(click, "echo")
    def test_key_deleted(self, echo: Mock):
        # given
        key_id = "key-1"

        self._request_wrapper.make_request_v2.return_value = MonolithResponse(
            data=Box({"deleted": True}), errors=None
        )

        # when
        self._service.delete(key_id)

        # then
        self._request_wrapper.make_request_v2.assert_called_once_with(
            query=Queries.delete.query,
            variables={"keyId": key_id},
            operation=Queries.delete.operation,
        )

        echo.assert_called_once_with("Key has been deleted.")

    @patch.object(click, "echo")
    def test_key_not_deleted(self, echo: Mock):
        # given
        key_id = "key-1"

        self._request_wrapper.make_request_v2.return_value = MonolithResponse(
            data=Box({"deleted": False}), errors=None
        )

        # when
        self._service.delete(key_id)

        # then
        self._request_wrapper.make_request_v2.assert_called_once_with(
            query=Queries.delete.query,
            variables={"keyId": key_id},
            operation=Queries.delete.operation,
        )

        echo.assert_called_once_with("Key was not deleted.")

    @patch.object(click, "echo")
    def test_get_all(self, echo: Mock):
        self._request_wrapper.make_request_v2.return_value = MonolithResponse(
            data=BoxList(
                [
                    {
                        "id": "key-id",
                        "description": "test key",
                        "scope": "spark",
                        "createdTime": "2021-01-01T00:00:00Z",
                        "createdBy": {
                            "id": "user-id",
                            "firstName": "Bilbo",
                            "lastName": "Baggins",
                            "email": "bilbo@theshire.com",
                        },
                    }
                ]
            ),
            errors=None,
        )

        # when
        self._service.get_all()

        # then
        self._request_wrapper.make_request_v2.assert_called_once_with(
            query=Queries.get_all.query, operation=Queries.get_all.operation
        )

        echo.assert_called_once_with(
            "╒════════╤═══════════════╤═════════╤══════════════════════╤═══════════════╕\n"
            "│ Id     │ Description   │ Scope   │ Created              │ Created By    │\n"
            "╞════════╪═══════════════╪═════════╪══════════════════════╪═══════════════╡\n"
            "│ key-id │ test key      │ spark   │ 2021-01-01T00:00:00Z │ Bilbo Baggins │\n"
            "╘════════╧═══════════════╧═════════╧══════════════════════╧═══════════════╛"
        )



import json
import os
from typing import Optional, Dict

import click
from tabulate import tabulate

from montecarlodata.common.data import (
    ConnectionOptions,
    ValidationResult,
    OnboardingConfiguration,
)
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, prompt_connection
from montecarlodata.integrations.onboarding.fields import (
    S3_CERT_MECHANISM,
    GQL_TO_FRIENDLY_CONNECTION_MAP,
    DATA_LAKE_WAREHOUSE_TYPE,
    EXPECTED_ADD_CONNECTION_RESPONSE_FIELD,
    CONFIRM_CONNECTION_VERBIAGE,
    VALIDATIONS_FAILED_VERBIAGE,
    SKIP_ADD_CONNECTION_VERBIAGE,
    CONNECTION_TEST_SUCCESS_VERBIAGE,
    CONNECTION_TEST_FAILED_VERBIAGE,
    ADD_CONNECTION_FAILED_VERBIAGE,
    ADD_CONNECTION_SUCCESS_VERBIAGE,
    MAIN_CONNECTION_TYPES,
)
from montecarlodata.queries.onboarding import ADD_CONNECTION_MUTATION
from montecarlodata.utils import GqlWrapper, AwsClientWrapper


class BaseOnboardingService:
    _VALIDATION_STATUS_HEADERS = ["Validation", "Result", "Details"]
    _VALIDATION_PASSED = "Passed"
    _VALIDATION_FAILED = "Failed"

    def __init__(
        self,
        config: Config,
        user_service: Optional[UserService] = None,
        request_wrapper: Optional[GqlWrapper] = None,
        aws_wrapper: Optional[AwsClientWrapper] = None,
    ):

        self._abort_on_error = True  # Aborts methods with deco on unhandled error
        self._dc_outputs = None

        self._user_service = user_service or UserService(config=config)
        self._request_wrapper = request_wrapper or GqlWrapper(config)
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(
            profile_name=config.aws_profile, region_name=config.aws_region
        )

    def onboard(self, **kwargs):
        """
        Convenience wrapper to validate and add a connection.
        """
        onboarding_config = OnboardingConfiguration(
            **kwargs
        )  # not passed as arg to avoid having to change all clients
        onboarding_config.connection_options = self._build_connection_options(
            **onboarding_config.connection_options
        )

        onboarding_config.warehouse_name = (
            onboarding_config.connection_options.monolith_base_payload.pop(
                "warehouseName", onboarding_config.connection_type
            )
        )
        # Prefer explicitly set types.
        if not onboarding_config.warehouse_type:
            onboarding_config.warehouse_type = (
                onboarding_config.connection_options.monolith_base_payload.pop(
                    "warehouseType", DATA_LAKE_WAREHOUSE_TYPE
                )
            )

        result = self._validate_connection(
            query=onboarding_config.validation_query,
            response_field=onboarding_config.validation_response,
            **onboarding_config.connection_options.monolith_base_payload,
        )

        if not onboarding_config.connection_options.validate_only:
            prompt = (
                VALIDATIONS_FAILED_VERBIAGE
                if result.has_warnings
                else CONFIRM_CONNECTION_VERBIAGE
            )
            prompt_connection(
                message=prompt,
                skip_prompt=onboarding_config.connection_options.auto_yes,
            )
            self._add_connection(
                temp_path=result.credentials_key, onboarding_config=onboarding_config
            )
        else:
            click.echo(SKIP_ADD_CONNECTION_VERBIAGE)

    def handle_cert(self, cert_prefix: str, options: Dict) -> None:
        """
        Handles cert payload from either an s3 path or file. Uploading the latter
        Options is updated if successful.
        """
        if options.get("cert_file") is not None:
            bucket_name = self._get_dc_property(prop="PrivateS3BucketArn").split(":")[
                5
            ]  # get name from arn
            object_name = os.path.join(
                cert_prefix, os.path.basename(options["cert_file"])
            )
            self._aws_wrapper.upload_file(
                bucket_name=bucket_name,
                object_name=object_name,
                file_path=options["cert_file"],
            )

            click.echo(
                f"Uploaded '{options['cert_file']}' to s3://{bucket_name}/{object_name}"
            )
            options["cert_s3"] = object_name

        ssl_options = {}
        if options.get("cert_s3") is not None:
            # reformat to generic options and specify a mechanism
            ssl_options = {
                "mechanism": S3_CERT_MECHANISM,
                "cert": options.pop("cert_s3"),
            }
        options.pop("cert_file", None)

        if options.get("skip_cert_verification") is not None:
            ssl_options["skip_verification"] = options.pop("skip_cert_verification")

        if ssl_options:
            options["ssl_options"] = ssl_options

    def _validate_connection(
        self, query: str, response_field: str, **kwargs
    ) -> ValidationResult:
        """
        Validate connection before adding using the expected gql response_field (e.g. testPrestoCredentials)
        """
        response = self._request_wrapper.make_request_v2(
            query=query, operation=response_field, variables=kwargs
        )

        summary = self._build_validation_summary(response.data)
        temp_path = response.data.get("key")
        warnings = response.data.get("warnings")

        if temp_path is not None:
            click.echo(f"{CONNECTION_TEST_SUCCESS_VERBIAGE}{summary}")
            return ValidationResult(
                has_warnings=warnings is not None and len(warnings) > 0,
                credentials_key=temp_path,
            )

        complain_and_abort(f"{CONNECTION_TEST_FAILED_VERBIAGE}{summary}")

    def _build_validation_summary(
        self, response: Dict, table_format: Optional[str] = "fancy_grid"
    ) -> str:
        data = []
        if response.get("validations"):
            for passed in response.get("validations"):
                data.append([passed.get("message"), self._VALIDATION_PASSED, None])

        if response.get("warnings"):
            for failed in response.get("warnings", []):
                data.append(
                    [
                        failed.get("message"),
                        self._VALIDATION_FAILED,
                        failed.get("data", {}).get("error"),
                    ]
                )

        if data:
            return "\n" + tabulate(
                data, headers=self._VALIDATION_STATUS_HEADERS, tablefmt=table_format
            )

        # do not provide a summary if no detailed response is available
        return ""

    def _add_connection(
        self, temp_path: str, onboarding_config: OnboardingConfiguration
    ) -> bool:
        """
        Add connection and setup any associated jobs
        """
        connection_request = {
            "key": temp_path,
            "connectionType": onboarding_config.connection_type,
        }
        connection_request.update(self._disambiguate_warehouses(onboarding_config))

        # Set optional properties
        if onboarding_config.warehouse_name:
            connection_request["name"] = onboarding_config.warehouse_name
        if onboarding_config.job_types:
            connection_request["jobTypes"] = onboarding_config.job_types
        if onboarding_config.job_limits:
            connection_request["jobLimits"] = json.dumps(onboarding_config.job_limits)
        if (
            onboarding_config.connection_options
            and onboarding_config.connection_options.dc_id
        ):
            connection_request["dcId"] = onboarding_config.connection_options.dc_id

        response = self._request_wrapper.make_request_v2(
            query=onboarding_config.connection_query or ADD_CONNECTION_MUTATION,
            operation=onboarding_config.connection_response
            or EXPECTED_ADD_CONNECTION_RESPONSE_FIELD,
            variables=connection_request,
        )

        connection_id = response.data.get("connection", {}).get("uuid")
        if connection_id is not None:
            click.echo(
                f"{ADD_CONNECTION_SUCCESS_VERBIAGE}"
                f"{GQL_TO_FRIENDLY_CONNECTION_MAP.get(onboarding_config.connection_type, onboarding_config.connection_type.capitalize())}."
            )
            return True
        complain_and_abort(ADD_CONNECTION_FAILED_VERBIAGE)

    def _disambiguate_warehouses(
        self, onboarding_config: OnboardingConfiguration
    ) -> Dict:
        """
        Determine type of connection request to build using the following criteria -
            1) If it is a main connection type or a traditional warehouse (i.e. not a lake), create a new warehouse.
            2) If a name has been passed, create the connection in that warehouse.
            3) If no name has been passed but there is only one warehouse, use it.
            4) Otherwise fail.
        """
        warehouse_type = (
            onboarding_config.warehouse_type or DATA_LAKE_WAREHOUSE_TYPE
        )  # default to data-lake
        connection_type = onboarding_config.connection_type
        create_warehouse = onboarding_config.create_warehouse

        # Main connections require the creation of a new warehouse.
        if (
            connection_type in MAIN_CONNECTION_TYPES
            or warehouse_type != DATA_LAKE_WAREHOUSE_TYPE
        ) and create_warehouse:
            return {"createWarehouseType": warehouse_type}

        name = onboarding_config.warehouse_name

        dw_id = self._get_warehouse_id_from_name(name)
        return {"dwId": dw_id}

    def _get_warehouse_id_from_name(self, name: Optional[str]) -> str:
        warehouses = self._user_service.warehouses

        # If it is not a new warehouse, find the warehouse with the given name to create the new connection belonging
        # to it. If the name is not found, it is an error.
        if name:
            warehouses = [
                warehouse for warehouse in warehouses if warehouse.get("name") == name
            ]
            if warehouses:
                return warehouses[0]["uuid"]
            complain_and_abort("Could not find warehouse with given name")
        # Name is optional if there is only one warehouse.
        elif len(warehouses) == 1:
            return warehouses[0]["uuid"]
        # If more than one warehouse exists, name is required.
        else:
            complain_and_abort("Name is required to disambiguate lake warehouses")

    def _get_bi_container_id_from_connection_id(
        self, bi_connection_id: Optional[str]
    ) -> str:
        """
        Exchange a Connection UUID for the parent BI Container's UUID
        """
        bi_containers = self._user_service.bi_containers
        for bi_container in bi_containers:
            connections = bi_container.get("connections")
            for connection in connections:
                if connection["uuid"] == bi_connection_id:
                    return bi_container["uuid"]
        complain_and_abort("Could not find BI container with given connection ID")

    def _get_dc_property(self, prop: str) -> Optional[str]:
        """
        Retrieve property from DC stack outputs
        """
        self._dc_outputs = self._dc_outputs or self._aws_wrapper.get_stack_outputs(
            self._user_service.active_collector["stackArn"]
        )  # cache lookup
        for output in self._dc_outputs:
            if output["OutputKey"] == prop:
                return output["OutputValue"]

    @staticmethod
    def _build_connection_options(**kwargs) -> ConnectionOptions:
        """
        Create connection options from arguments
        """
        connection_options = ConnectionOptions(**kwargs)

        if connection_options.monolith_connection_payload:
            connection_options.monolith_base_payload[
                "connectionOptions"
            ] = connection_options.monolith_connection_payload
        return connection_options



CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE = """
mutation createOrUpdateMonteCarloConfigTemplate($namespace: String!, $configTemplateJson: String!, 
    $dryRun: Boolean, $misconfiguredAsWarning: Boolean,  $resource: String) {
  createOrUpdateMonteCarloConfigTemplate(
    configTemplateJson: $configTemplateJson,
    namespace: $namespace,
    dryRun: $dryRun,
    misconfiguredAsWarning: $misconfiguredAsWarning,
    resource: $resource            
  ) {
    response {
      resourceModifications {
        type
        description
        resourceAsJson
      }
      changesApplied
      errorsAsJson
      warningsAsJson
    }
  }
}
"""

DELETE_MONTE_CARLO_CONFIG_TEMPLATE = """
mutation deleteMonteCarloConfigTemplate($namespace: String!, $dryRun: Boolean) {
  deleteMonteCarloConfigTemplate(
    namespace: $namespace,
    dryRun: $dryRun
  ) {
    response {
      changesApplied
      numDeleted
    }
  }
}
"""



import copy
from unittest import TestCase
from unittest.mock import Mock, patch

import click
from box import Box

from montecarlodata.collector.fields import ADD_DC_PROMPT_VERBIAGE
from montecarlodata.collector.management import CollectorManagementService
from montecarlodata.common.data import MonolithResponse
from montecarlodata.common.user import UserService
from montecarlodata.queries.collector import ADD_COLLECTOR_RECORD
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.helpers import capture_function
from tests.test_common_user import _SAMPLE_CONFIG

_SAMPLE_STACK_ARN = "stack_arn"
_SAMPLE_GATEWAY_ID = "gateway_id"
_SAMPLE_TEMPLATE_URI = (
    "https://s3.amazonaws.com/bucket/public/customer_templates/id.json"
)
_SAMPLE_INVALID_LAUNCH_URL = f"https://test.com/cf/home?region=us-east-1#/stacks/create/review?stackName=monte-carlo&malformed"
_SAMPLE_VALID_LAUNCH_URL = f"https://test.com/cf/home?region=us-east-1#/stacks/create/review?stackName=monte-carlo&templateURL={_SAMPLE_TEMPLATE_URI}"
_SAMPLE_COLLECTOR_RESPONSE = {
    "generateCollectorTemplate": {
        "dc": {
            "templateLaunchUrl": _SAMPLE_VALID_LAUNCH_URL,
            "stackArn": _SAMPLE_STACK_ARN,
            "active": True,
            "apiGatewayId": _SAMPLE_GATEWAY_ID,
            "templateVariant": "dionysus",
        }
    }
}
_SAMPLE_EXTRACTED_COLLECTOR_RESPONSE = _SAMPLE_COLLECTOR_RESPONSE[
    "generateCollectorTemplate"
]
_SAMPLE_MONOLITH_RESPONSE = MonolithResponse(data=_SAMPLE_EXTRACTED_COLLECTOR_RESPONSE)
_SAMPLE_COLLECTORS_RESPONSE = [
    {
        "uuid": "1234",
        "stackArn": "aws-loki",
        "active": True,
        "customerAwsAccountId": "test",
        "templateProvider": "cloudformation",
        "templateVariant": "dionysus",
        "templateVersion": "42",
        "codeVersion": "2042",
        "lastUpdated": None,
    },
    {
        "uuid": "5678",
        "stackArn": "aws-thor",
        "active": True,
        "templateProvider": "terraform",
        "templateVariant": "foo",
        "templateVersion": "42",
        "codeVersion": "2042",
        "lastUpdated": None,
    },
    {
        "uuid": "0912",
        "stackArn": "aws-odin",
        "active": False,
        "templateProvider": None,
        "templateVariant": None,
        "templateVersion": "42",
        "codeVersion": "2042",
        "lastUpdated": None,
    },
]


class CollectorManagementTest(TestCase):
    def setUp(self) -> None:
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)
        self._user_service_mock = Mock(autospec=UserService)

        self._service = CollectorManagementService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock,
        )

    def test_get_template_with_response(self):
        self._request_wrapper_mock.make_request_v2.return_value = None
        with self.assertRaises(click.exceptions.Abort):
            self._service.echo_template()

    def test_get_template_with_invalid_link(self):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"dc": {"templateLaunchUrl": _SAMPLE_INVALID_LAUNCH_URL}}
        )

        with self.assertRaises(click.exceptions.Abort):
            self._service.echo_template()

    def test_get_template_with_valid_link(self):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"dc": {"templateLaunchUrl": _SAMPLE_VALID_LAUNCH_URL}}
        )

        std_out = capture_function(function=self._service.echo_template).std_out
        self.assertEqual(std_out.getvalue().strip(), _SAMPLE_TEMPLATE_URI)

    def test_upgrade_with_inactive_collector(self):
        sample_collector_response = copy.deepcopy(_SAMPLE_COLLECTOR_RESPONSE)
        sample_collector_response["generateCollectorTemplate"]["dc"]["active"] = False

        self._request_wrapper_mock.make_request.return_value = sample_collector_response
        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

    def test_upgrade_with_missing_fields(self):
        self._request_wrapper_mock.make_request.return_value = {
            "generateCollectorTemplate": {"dc": {}}
        }
        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

    def test_upgrade(self):
        self._aws_wrapper_mock.get_stack_parameters.return_value = None
        self._aws_wrapper_mock.upgrade_stack.return_value = True

        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        self._service.upgrade_template()

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN
        )
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN,
            template_link=_SAMPLE_TEMPLATE_URI,
            parameters=None,
        )
        self._aws_wrapper_mock.deploy_gateway.assert_called_once_with(
            gateway_id=_SAMPLE_GATEWAY_ID
        )

    def test_upgrade_to_janus(self):
        self._aws_wrapper_mock.get_stack_parameters.return_value = [
            {"ParameterKey": "CreateCrossAccountRole", "ParameterValue": "True"},
            {"ParameterKey": "CreateEventInfra", "ParameterValue": "True"},
            {"ParameterKey": "EnableGatewayAccessLogging", "ParameterValue": "False"},
            {"ParameterKey": "ExistingS3Vpce", "ParameterValue": "True"},
            {"ParameterKey": "ExistingSubnetIds", "ParameterValue": "True"},
            {"ParameterKey": "ExistingVpcId", "ParameterValue": "True"},
            {"ParameterKey": "subnetCIDRs", "ParameterValue": "2222"},
            {"ParameterKey": "vpcCIDR", "ParameterValue": "1111"},
            {
                "ParameterKey": "workerLambdaConcurrentExecutions",
                "ParameterValue": "12",
            },
        ]
        self._aws_wrapper_mock.upgrade_stack.return_value = True

        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        self._service.upgrade_template(
            update_infra=True,
            new_params={"EnableRemoteUpdates": "True", "CreateEventInfra": "False"},
        )

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN
        )
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN,
            template_link=_SAMPLE_TEMPLATE_URI,
            parameters=[
                # updated parameter
                {
                    "ParameterKey": "CreateEventInfra",
                    "ParameterValue": "False",
                    "UsePreviousValue": False,
                },
                # preserved parameters
                {"ParameterKey": "ExistingS3Vpce", "UsePreviousValue": True},
                {"ParameterKey": "ExistingSubnetIds", "UsePreviousValue": True},
                {"ParameterKey": "ExistingVpcId", "UsePreviousValue": True},
                # migrated parameters
                {"ParameterKey": "SubnetCIDRs", "ParameterValue": "2222"},
                {"ParameterKey": "VpcCIDR", "ParameterValue": "1111"},
                {
                    "ParameterKey": "WorkerLambdaConcurrentExecutions",
                    "ParameterValue": "12",
                },
                # new parameter
                {"ParameterKey": "EnableRemoteUpdates", "ParameterValue": "True"},
            ],
        )
        self._aws_wrapper_mock.deploy_gateway.assert_called_once_with(
            gateway_id=_SAMPLE_GATEWAY_ID
        )

    def test_failed_stack_upgrade(self):
        self._aws_wrapper_mock.get_stack_parameters.return_value = None
        self._aws_wrapper_mock.upgrade_stack.return_value = False

        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN
        )
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN,
            template_link=_SAMPLE_TEMPLATE_URI,
            parameters=None,
        )
        self._aws_wrapper_mock.deploy_gateway.assert_not_called()

    def test_deploy(self):
        stack_name, termination_protection = "foo", False
        collector_response = copy.deepcopy(_SAMPLE_COLLECTOR_RESPONSE)
        collector_response["generateCollectorTemplate"]["dc"]["active"] = False
        monolith_response = MonolithResponse(
            data=collector_response["generateCollectorTemplate"]
        )

        self._aws_wrapper_mock.create_stack.return_value = True

        self._request_wrapper_mock.make_request_v2.return_value = monolith_response
        self._service.deploy_template(
            stack_name=stack_name, termination_protection=termination_protection
        )

        self._aws_wrapper_mock.create_stack.assert_called_once_with(
            stack_name=stack_name,
            template_link=_SAMPLE_TEMPLATE_URI,
            termination_protection=termination_protection,
            parameters=[],
        )

    def test_deploy_with_an_active_collector(self):
        stack_name, termination_protection = "foo", True
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        with self.assertRaises(click.exceptions.Abort):
            self._service.deploy_template(
                stack_name=stack_name, termination_protection=termination_protection
            )

    def test_replace_params(self):
        existing_params, new_params = [
            {"ParameterKey": "foo", "ParameterValue": "6"}
        ], {"foo": "42"}
        parameters = self._service._build_param_list(
            existing_params=existing_params, new_params=new_params
        )
        self.assertEqual(
            parameters,
            [
                {
                    "ParameterKey": "foo",
                    "ParameterValue": "42",
                    "UsePreviousValue": False,
                }
            ],
        )

    def test_only_existing_params(self):
        existing_params = [{"ParameterKey": "foo", "ParameterValue": "42"}]
        parameters = self._service._build_param_list(
            existing_params=existing_params, new_params=None
        )
        self.assertEqual(
            parameters, [{"ParameterKey": "foo", "UsePreviousValue": True}]
        )

    def test_completely_new_params(self):
        parameters = self._service._build_param_list(
            existing_params=[], new_params={"foo": "42"}
        )
        self.assertEqual(parameters, [{"ParameterKey": "foo", "ParameterValue": "42"}])

    @patch("montecarlodata.collector.management.click")
    def test_get_launch_link(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        self._service.launch_quick_create_link(dry=False)

        click_mock.launch.assert_called_once_with(_SAMPLE_VALID_LAUNCH_URL)
        click_mock.echo.assert_not_called()

    @patch("montecarlodata.collector.management.click")
    def test_get_launch_link_dry_run(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        self._service.launch_quick_create_link(dry=True)

        click_mock.echo.assert_called_once_with(_SAMPLE_VALID_LAUNCH_URL)
        click_mock.launch.assert_not_called()

    def test_set_region(self):
        region = "loki"
        service = CollectorManagementService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            aws_region_override=region,
        )
        self.assertEqual(service._collector_region, region)
        self.assertEqual(self._service._collector_region, "us-east-1")
        self.assertEqual(self._service._collector_region, _SAMPLE_CONFIG.aws_region)

    @patch("montecarlodata.collector.management.click")
    def test_echo_collectors(self, click_mock):
        self._user_service_mock.collectors = _SAMPLE_COLLECTORS_RESPONSE

        self._service.echo_collectors(table_format="plain")
        click_mock.echo.assert_called_once_with(
            "AWS Stack ARN      ID    Version  Template                    Last updated    Active\n"
            "aws-loki         1234       2042  cloudformation:dionysus:42  -               True\n"
            "aws-thor         5678       2042  terraform:foo:42            -               True\n"
            "aws-odin         0912       2042  cloudformation:July-2019    -               False"
        )

    @patch("montecarlodata.collector.management.click")
    def test_echo_collectors_with_only_active(self, click_mock):
        self._user_service_mock.collectors = _SAMPLE_COLLECTORS_RESPONSE

        self._service.echo_collectors(table_format="plain", active_only=True)
        click_mock.echo.assert_called_once_with(
            "AWS Stack ARN      ID    Version  Template                    Last updated    Active\n"
            "aws-loki         1234       2042  cloudformation:dionysus:42  -               True\n"
            "aws-thor         5678       2042  terraform:foo:42            -               True"
        )

    @patch.object(CollectorManagementService, "launch_quick_create_link")
    @patch("montecarlodata.collector.management.prompt_connection")
    @patch("montecarlodata.collector.management.click")
    def test_add_collector(
        self, click_mock, prompt_connection_mock, launch_quick_create_link_mock
    ):
        mocked_region = "us-east-1"
        click_mock.prompt.return_value = mocked_region
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data=Box({"dc": {"uuid": "1234"}})
        )

        self._service.add_collector()

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_COLLECTOR_RECORD, operation="createCollectorRecord"
        )
        click_mock.echo.assert_called_once_with(
            "Created collector record with ID '1234'"
        )
        prompt_connection_mock.assert_called_once_with(message=ADD_DC_PROMPT_VERBIAGE)
        launch_quick_create_link_mock.assert_called_once_with(
            dry=False, dc_id="1234", collection_region=mocked_region
        )

    @patch("montecarlodata.collector.management.click")
    def test_add_collector_no_prompt(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data=Box({"dc": {"uuid": "1234"}})
        )

        self._service.add_collector(no_prompt=True)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_COLLECTOR_RECORD,
            operation="createCollectorRecord",
        )
        click_mock.echo.assert_called_once_with(
            "Created collector record with ID '1234'"
        )



import os
from typing import Optional

from pycarlo.core import Client, Mutation, Query
import click
from tabulate import tabulate

from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.common.data import OnboardingConfiguration
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    DATABRICKS_DELTA_CONNECTION_TYPE,
    DATABRICKS_SQL_WAREHOUSE_CONNECTION_TYPE,
    DATA_LAKE_WAREHOUSE_TYPE,
    EXPECTED_DATABRICKS_GQL_RESPONSE_FIELD,
    EXPECTED_DATABRICKS_SQL_WAREHOUSE_GQL_RESPONSE_FIELD,
)
from montecarlodata.queries.onboarding import (
    TEST_DATABRICKS_CRED_MUTATION,
    DatabricksSqlWarehouseOnboardingQueries,
)

DEFAULT_GATEWAY_URL = os.getenv(
    "MC_INTEGRATION_GATEWAY_URL",
    "https://integrations.getmontecarlo.com/databricks/metadata",
)
# Secret scope and name defined in the data collector:
# https://github.com/monte-carlo-data/data-collector/blob/e5db61eacd10e0388acd9ff8a32fe60644b36797/lambdas/api_lambdas/execute_databricks_actions.py#L61  # noqa
DEFAULT_SECRET_SCOPE = "monte-carlo-collector-gateway-scope"
DEFAULT_SECRET_NAME = "monte-carlo-collector-gateway-secret"


class DatabricksOnboardingService(BaseOnboardingService):

    _DATABRICKS_CONFIG_KEYS = [
        "databricks_workspace_url",
        "databricks_workspace_id",
        "databricks_cluster_id",
        "databricks_token",
    ]

    def __init__(self, config: Config, mc_client: Optional[Client] = None, **kwargs):
        super().__init__(config, **kwargs)
        self._mc_client = mc_client or Client()

    @manage_errors
    def onboard_databricks_metastore(self, connection_type, **kwargs) -> None:
        """
        Onboard a databricks metastore connection
        """

        validation_query = Query()
        validation_query.validate_connection_type(
            warehouse_type=DATA_LAKE_WAREHOUSE_TYPE, connection_type=connection_type
        )
        self._mc_client(validation_query)

        databricks_config = {key: kwargs[key] for key in self._DATABRICKS_CONFIG_KEYS}

        onboarding_config = OnboardingConfiguration(
            connection_type=connection_type, **kwargs
        )
        connection_options = self._build_connection_options(
            **onboarding_config.connection_options
        )

        # Create secret
        if not kwargs.get("skip_secret_creation", False):
            mutation = Mutation()
            mutation.create_databricks_secret(
                databricks_config=databricks_config,
                secret_name=kwargs["databricks_secret_key"],
                scope_name=kwargs["databricks_secret_scope"],
                connection_options=connection_options.monolith_connection_payload,
            )
            self._mc_client(mutation)

        # Create notebook
        if not kwargs.get("skip_notebook_creation", False):
            mutation = Mutation()
            mutation.create_databricks_notebook_job(
                databricks_config=databricks_config,
                connection_options=connection_options.monolith_connection_payload,
            )
            resp = self._mc_client(mutation)
            databricks_job_config = resp.create_databricks_notebook_job.databricks
            job_info = dict(
                databricks_job_id=databricks_job_config.workspace_job_id,
                databricks_job_name=databricks_job_config.workspace_job_name,
                databricks_notebook_path=databricks_job_config.workspace_notebook_path,
                databricks_notebook_source=databricks_job_config.notebook_source,
                databricks_notebook_version=databricks_job_config.notebook_version,
            )
        else:
            job_info = dict(
                databricks_job_id=kwargs.pop("databricks_job_id"),
                databricks_job_name=kwargs.pop("databricks_job_name"),
                databricks_notebook_path=kwargs.pop("databricks_notebook_path"),
                databricks_notebook_source=kwargs.pop("databricks_notebook_source"),
            )

        project_info = {}

        if connection_type == DATABRICKS_DELTA_CONNECTION_TYPE:
            project_query = Query()
            project_query.get_projects(
                warehouse_type=DATA_LAKE_WAREHOUSE_TYPE,
            )
            project_result = self._mc_client(project_query)

            project_id = project_result.get_projects.projects[0]
            project_info = {"project_id": project_id}

        job_limits = {
            **dict(
                integration_gateway=dict(
                    gateway_url=DEFAULT_GATEWAY_URL,
                    databricks_secret_key=kwargs["databricks_secret_key"],
                    databricks_secret_scope=kwargs["databricks_secret_scope"],
                )
            ),
            **project_info,
            **job_info,
        }

        # Onboard
        self.onboard(
            validation_query=TEST_DATABRICKS_CRED_MUTATION,
            validation_response=EXPECTED_DATABRICKS_GQL_RESPONSE_FIELD,
            connection_type=connection_type,
            job_limits=job_limits,
            **kwargs
        )

    @manage_errors
    def update_databricks_notebook(self, **kwargs):
        mutation = Mutation()
        mutation.update_databricks_notebook(**kwargs)
        resp = self._mc_client(mutation)
        click.echo(resp.update_databricks_notebook)

    @manage_errors
    def get_databricks_job_info(self, **kwargs):
        query = Query()
        query.get_databricks_metadata_job_info(**kwargs)
        resp = self._mc_client(query)
        attributes = [
            "workspace_job_id",
            "workspace_job_name",
            "workspace_notebook_path",
            "notebook_source",
            "notebook_version",
        ]
        info_list = [
            {attribute: getattr(info, attribute) for attribute in attributes}
            for info in resp.get_databricks_metadata_job_info
        ]
        click.echo(tabulate(info_list, headers="keys"))

    @manage_errors
    def get_current_databricks_notebook_version(self):
        query = Query()
        query.get_current_databricks_notebook_version()
        resp = self._mc_client(query)
        click.echo(resp.get_current_databricks_notebook_version)

    @manage_errors
    def onboard_databricks_sql_warehouse(self, **kwargs):
        # Onboard
        self.onboard(
            validation_query=DatabricksSqlWarehouseOnboardingQueries.test_credentials.query,
            validation_response=EXPECTED_DATABRICKS_SQL_WAREHOUSE_GQL_RESPONSE_FIELD,
            connection_type=DATABRICKS_SQL_WAREHOUSE_CONNECTION_TYPE,
            **kwargs
        )



import io
import sys
from dataclasses import dataclass
from typing import Callable, Dict, Optional, Any


@dataclass
class FunctionComponents:
    std_out: io.StringIO
    exception: Optional[Any] = None


def capture_function(
    function: Callable, params: Optional[Dict] = None
) -> FunctionComponents:
    """
    Helper function to capture standard out and any exceptions from a function
    """
    exception = None
    std_out = io.StringIO()

    sys.stdout = std_out
    try:
        function(**(params or {}))
    except Exception as err:
        exception = err
    sys.stdout = sys.__stdout__
    return FunctionComponents(std_out=std_out, exception=exception)
